///
/// This file is part of Rheolef.
///
/// Copyright (C) 2000-2009 Pierre Saramito <Pierre.Saramito@imag.fr>
///
/// Rheolef is free software; you can redistribute it and/or modify
/// it under the terms of the GNU General Public License as published by
/// the Free Software Foundation; either version 2 of the License, or
/// (at your option) any later version.
///
/// Rheolef is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/// GNU General Public License for more details.
///
/// You should have received a copy of the GNU General Public License
/// along with Rheolef; if not, write to the Free Software
/// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
///
/// =========================================================================
/*Prog:proj
NAME: @code{proj} - field projection 
@pindex proj
@cindex RHEOPATH environment variable
@cindex projection
@clindex field
@apindex P1
SYNOPSIS:
   @example
       proj @var{filename}
   @end example
DESCRIPTION:       
   Project a finite-element field on a P1 finite element space.
EXAMPLE: 
   Enter command as:
   @example
    	proj my-p0-result.field > my-p1-result.field
   @end example
   or
   @example
    	proj - < my-p0-result.field > my-p1-result.field
   @end example
FILE OPTIONS:
   @table @code
   @itemx -I@var{geodir}
   	add 'geodir' to the @samp{RHEOPATH} search path.
    	See also geo for RHEOPATH mechanism. 
   @itemx @var{filename}
   	specifies the name of the file containing
        the input field.
   @itemx -
   	read field on standard input instead on a file.
   @item -output-as-double
        Avoid variation of output float formats
        when using high precision @code{doubledouble} or
        @code{bigfloat} classes.
        This option is used for non-regression test purpose.
   @item -quadrature
	Use the mass lumping 
        diagonalization procedure, instead of the default
	exact mass matrix.
   @item -exact
	Use exact mass matrix instead of the mass lumping 
        diagonalization procedure. This is the default.
   @item -null-bdr
	The P1-continuous fonction will have zero
	value on boundary domains.
@findex catchmark iostream manipulator
   @item -name @var{string}
        The field is marked with the corresponding name
	by using the @code{catchmark} iostream manipulator.
	This option is usefull when using a multi-field input.
   @end table
BUG:
   Field is printed in simple precision.
AUTHOR: 
    LMC-IMAG, 38041 Grenoble cedex 9, France
    | Pierre.Saramito@imag.fr
SEE ALSO:
    "field"(1), "space"(1), "geo"(3), "mass"(7)
DATE:
    23 march 1998
End:
*/
#include "rheolef/rheolef.h"
using namespace std;

void usage()
{
      cerr << "proj: usage: proj"
	   << " {-Igeodir}*"
	   << " [-name string]"
	   << " [-output-as-double]"
	   << " [-exact]"
	   << " [-quadrature]"
	   << " [-null-bdr]"
	   << " - | file[.field]"
	   << endl;
      cerr << "example:\n";
      cerr << "  proj -I../geo result.field\n";
      exit (1);
}
int main(int argc, char**argv)
{
    if (argc <= 1) usage();
    //
    // set default options
    //
    bool exact = true;
    bool null_bdr = false;
    int digits10 = numeric_limits<Float>::digits10;
    string field_name;
    //
    // load geometry
    //
    field u;
    string name;
    bool have_input = false;

    for (int i = 1; i < argc; i++) {

	if      (strcmp ("-exact",argv [i]) == 0)    exact = true;
	else if (strcmp ("-quadrature",argv [i]) == 0) exact = false;
	else if (strcmp ("-null-bdr",argv [i]) == 0) null_bdr = true;
	else if (strcmp ("-name",argv [i]) == 0) field_name = argv[++i];
        else if (strcmp ("-output-as-double", argv[i]) == 0) {

	    numeric_flags<Float>::output_as_double(true);

        } else if (argv [i][0] == '-' && argv [i][1] == 'I') {

            append_dir_to_rheo_path (argv[i]+2);

        } else if (strcmp (argv [i], "-") == 0) {
	    
	    // input field on standard input
            clog << "! load field on stdin" << endl;
	    cin >> u;
	    name = "output";
    	    have_input = true;
	
	} else if (argv [i][0] == '-') {
	    
	    usage();
        
	} else {

            // input on file
            name = get_basename(delete_suffix (delete_suffix (argv[i], "gz"), "field"));
            u = field (argv[i]);
    	    have_input = true;
	
	    // force "output" base name
	    // name = "output";
        }
    }
    if (!have_input) {
	cerr << "field: no input specifield." << endl;
	usage();
    }
    //
    // build the projection
    //
    space U = u.get_space();
    const geo& g = u.get_geo();
    space V(g, "P1", U.get_valued());
    if (null_bdr) {
	for (geo::size_type i = 0; i < g.n_domain(); i++) 
	    V.block(g.get_domain(i));
    }
    field v(V);
    form p(U, V, "mass");
    if (!exact) {
	form_diag m(V, "mass");
    	v.u = (1./m.uu)*(p.uu*u.u);
    } else {
	form m(V, V, "mass");
    	ssk<Float> fact = ldlt(m.uu);
        v.u = fact.solve((p*u).u);
    }
    v.b = 0;
    if (field_name != "") cout << catchmark(field_name);
    cout << v;
    return 0;
    //
    // then output field
    //
    cout << setbasename(name) 
	 << setprecision(digits10) 
	 << v;
    return 0;
}
