#ifndef _RHEO_CURL_H
#define _RHEO_CURL_H
///
/// This file is part of Rheolef.
///
/// Copyright (C) 2000-2009 Pierre Saramito <Pierre.Saramito@imag.fr>
///
/// Rheolef is free software; you can redistribute it and/or modify
/// it under the terms of the GNU General Public License as published by
/// the Free Software Foundation; either version 2 of the License, or
/// (at your option) any later version.
///
/// Rheolef is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/// GNU General Public License for more details.
///
/// You should have received a copy of the GNU General Public License
/// along with Rheolef; if not, write to the Free Software
/// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
/// 
/// =========================================================================

/*Class:curl
NAME: @code{curl} -- curl operator
@bfindex curl
@cindex curl operator
@apindex P1
@apindex P2
@apindex P1d
@apindex P0
SYNOPSIS:
    @example
      form(const space V, const space& M, "curl");
    @end example
DESCRIPTION:       
    @noindent
    Assembly the form associated to the
    curl operator on finite element space.
@tex  
    In three dimensions, both ${\bf V}$ and ${\bf M}$ are vector-valued:
    $$ 
      b({\bf u},{\bf q}) = \int_\Omega {\bf curl}\,{\bf u} . {\bf q} \, dx
    $$ 
    for all ${\bf u} \in {\bf V}$ and ${\bf q}\in {\bf M}$.

    In two dimensions, only ${\bf V}$ is vector-valued:
    $$ 
      b({\bf u},q) = \int_\Omega {\rm curl}\,{\bf u} \, q \, dx
    $$ 
    for all ${\bf u} \in {\bf V}$ and $q\in M$.

@end tex  
    
    @noindent
    The V space may be a either
    @samp{P2} finite element space,
    while the M space may be @samp{P1d}.
    See also @ref{form class} and @ref{space class}.

EXAMPLE:
    @noindent
    The following piece of code build the divergence form
    associated to the @samp{P2} approximation for a three dimensional
    geometry:
    @example
   	geo omega("cube");
   	space V(omega, "P2",  "vector");
   	space M(omega, "P1d", "vector");
   	form b(V, M, "curl");
    @end example
    while this code becomes in two dimension:
    @example
   	geo omega("square");
   	space V(omega, "P2",  "vector");
   	space M(omega, "P1d");
   	form b(V, M, "curl");
    @end example
AUTHOR: 
    LMC-IMAG, 38041 Grenoble cedex 9, France
    | Pierre.Saramito@imag.fr
DATE:
    3 march 2001
End:
*/

#include "rheolef/form_element_rep.h"

class curl: public form_element_rep {

public:
// allocator:

    curl() : form_element_rep() {}

// virtual accessor:

    void operator() (const geo_element& K, ublas::matrix<Float>& m) const;
    size_type n_derivative() const;
};
#endif // _RHEO_CURL_H
