/*
 * File that facilitates C++ program debugging.
 *
 * Copyright 2000 by Gray Watson
 *
 * This file is part of the dmalloc package.
 *
 * Permission to use, copy, modify, and distribute this software for
 * any purpose and without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all
 * copies, and that the name of Gray Watson not be used in advertising
 * or publicity pertaining to distribution of the document or software
 * without specific, written prior permission.
 *
 * Gray Watson makes no representations about the suitability of the
 * software described herein for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * The author may be contacted via http://dmalloc.com/
 *
 * $Id: dmallocc.cc,v 1.2 2008/09/24 09:35:22 saramito Exp $
 */

/*
 * This file is used to effectively redirect new to the more familiar
 * malloc and delete to the more familiar free so they can be debugged
 * with the debug malloc library..  They also give the known error
 * behavior, too.
 *
 * Compile and link this in with the C++ program you want to debug.
 *
 * NOTE: I am not a C++ hacker so feedback in the form of other hints
 * and ideas for C++ users would be much appreciated.
 */
 
extern "C" {
#include <stdlib.h>
#include <stdio.h>

#define DMALLOC_DISABLE

#include "dmalloc.h"
#include "return.h"
DMALLOC_PNT	_loc_malloc(const char *file, const int line,
			    const DMALLOC_SIZE size, const int calloc_b,
			    const DMALLOC_SIZE alignment);
int	_loc_free(const char *file, const int line, DMALLOC_PNT pnt);
}

// --------------------------------------------------------------------
/*
 * An overload function for the C++ new.
 */
void *
operator new(size_t size)
{
  char	*file;
  void *temp;
  GET_RET_ADDR(file);
  if (size == 0) return (void*)1;
  temp = _malloc_leap(file, 0, size);
  return temp;
}

/*
 * An overload function for the C++ new[].
 */
void *
operator new[](size_t size)
{
  char	*file;
  GET_RET_ADDR(file);
  if (size == 0) return (void*)1;
  return _malloc_leap(file, 0, size);
}

/*
 * An overload function for the C++ delete.
 */
void
operator delete(void *pnt)
{
  char	*file;
  GET_RET_ADDR(file);
  _free_leap(file, 0, pnt);
}

// --------------------------------------------------------------------
/*
 * An overload function for the C++ delete[].  Thanks to Jens Krinke
 * <j.krinke@gmx.de>
 */
void
operator delete[](void *pnt)
{
  char	*file;
  GET_RET_ADDR(file);
  _free_leap(file, 0, pnt);
}

/*
 * New versions of overloaded functions by James Hawtin oolon@ankh.org
 *
 * To all the C++ hackers out the I know I could have used x(x, y=0, z=0)
 * format to include the old and new versions in one however C++ linkers get a
 * bit confused and just use the standard ones rather than the dmalloc ones.
 */

/*
 * Ok I am using the loc version of the functions here because the first leap 
 * call is not correctly allocated a position in the program, because dmalloc
 * make the decision to call the malloc which the does the initialisation, so 
 * the second one works ?!?!?!?!
 */

void *operator new(size_t size, const char *const file, int line)
{
  if (size == 0) return (void*)1;
  return _loc_malloc(file, line, size, 0, 0);
}

void *operator new[](size_t size,  const char *const file, int line)
{
  if (size == 0) return (void*)1;
  return _loc_malloc(file, line, size, 0, 0);
}

// --------------------------------------------------------------------
class dmalloc_shutdown_call
{
public:
  ~dmalloc_shutdown_call();
};

extern "C"
{
  void _dmalloc_shutdown(void);
};

dmalloc_shutdown_call::~dmalloc_shutdown_call()
{
  _dmalloc_shutdown();
}

dmalloc_shutdown_call dmalloc_helper;
