///
/// This file is part of Rheolef.
///
/// Copyright (C) 2000-2009 Pierre Saramito <Pierre.Saramito@imag.fr>
///
/// Rheolef is free software; you can redistribute it and/or modify
/// it under the terms of the GNU General Public License as published by
/// the Free Software Foundation; either version 2 of the License, or
/// (at your option) any later version.
///
/// Rheolef is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/// GNU General Public License for more details.
///
/// You should have received a copy of the GNU General Public License
/// along with Rheolef; if not, write to the Free Software
/// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
///
/// =========================================================================
#include <sys/types.h> /* for pid_t */
#include <errno.h>     /* for errno, EINTR */
#include <stdio.h>
#include "rheolef/procbuf_c.h"
#include "rheolef/compiler.h"

#include <sys/types.h>

#  include <signal.h>
#  include <unistd.h>
#  include <stdlib.h>
#  include <sys/wait.h>

extern pid_t fork();
extern int   pipe();
extern int   dup2();

#define READS  4 /* Reading allowed */
#define WRITES 8 /* Writing allowed */

static struct procbuf_c*  proc_file_chain = 0;

struct procbuf_c*
proc_open (
    struct procbuf_c*   fp,
    const char *command, 
    const char *mode)
{
  if (!fp || !command || !mode) return 0;

  int p_desc[2];
  int p_desc1[2];
  int parent_desc, child_desc;
  int read_or_write;
  int status;
  pid_t pid;
  
  // fprintf (stderr, "proc_open: \"%s\"\n", command);

  /* p_desc[0] = pipe descriptor where parent can read	*/
  /* p_desc[1] = pipe descriptor where parent can write	*/
  status = pipe(p_desc);
  if (status != 0) {
    // fprintf (stderr, "proc_open: pipe status != 0 => RETURN 0");
    return 0;
  }
  if (strcmp(mode, "r+") == 0) {
      
      // fprintf (stderr, "proc_open read: READ | WRITE\n");

      // pipe0: parent read, child write
      parent_desc = p_desc[0];
      child_desc  = p_desc[1];
      read_or_write = READS | WRITES; 
      
      // pipe1: parent write, child read
      status = pipe(p_desc1);
      if (status != 0) {
        // fprintf (stderr, "proc_open: pipe status1 != 0 => RETURN 0");
        return 0;
      }
      // int parent_desc1 = p_desc1 [1];
      // int child_desc1  = p_desc1 [0];

  } else if (mode[0] == 'r') {

      // fprintf (stderr, "proc_open read: READ\n");
      parent_desc = p_desc[0];
      child_desc  = p_desc[1];
      read_or_write = READS; 
  
  } else if (mode[0] == 'w') {

      // fprintf (stderr, "proc_open read: WRITE\n");
      parent_desc = p_desc[1];
      child_desc  = p_desc[0];
      read_or_write = WRITES; 
  
  } else {
      return 0;
  }
  // fprintf (stderr, "proc_open: will start a fork...\n");
  pid = fp -> pid = fork();
  /*
   * the child work: he run the command (child_desc not used)
   * the parent get result (read) or put order(write) via parent_desc
   */
  if (pid == -1) {

      close(p_desc[0]);
      close(p_desc[1]);
      // fprintf (stderr, "unable to create a child process.\n");
      return 0;
  
  } else if (pid == 0) { /* child works... */


      int child_std_desc;
      /* 
       * redirect standard output/input on the pipe;
       * the next command will use pipe as input/output
       */
      if (read_or_write != (READS | WRITES)) {

          child_std_desc = ((mode[0] == 'r') ? 1 : 0);
          close(parent_desc);
          if (child_desc != child_std_desc) {

	      dup2(child_desc, child_std_desc);
	      close(child_desc);
          }
      } else {
          // fprintf (stderr, "CHILD: (in=%d,out=%d)\n", p_desc1 [0], p_desc [1]);

	  // child std input := parent output associate := p_desc1 [0]
	  if (p_desc1 [0] != 0) {
	      dup2(p_desc1 [0], 0);
	      close (p_desc1 [0]);
          }
          close(p_desc1 [1]);
	  
	  // child std output := parent input associate := p_desc1 [0]
	  if (p_desc[1] != 1) {
	      dup2(p_desc[1], 1);
	      close (p_desc[1]);
          }
          close(p_desc [0]);
      }
      /* Posix.2:  "popen() shall ensure that any streams from previous
         popen() calls that remain open in the parent process are closed
	 in the new child process." */

      while (proc_file_chain) {

	  close (proc_file_chain->fileno);
	  if (proc_file_chain->fileno1) 
	    close (proc_file_chain->fileno1);
	  proc_file_chain = proc_file_chain->next;
      }
      execl("/bin/sh", "sh", "-c", command, 0);
      /*
       * if return to the child process, an error has occured
       */
      // trace_macro ("pstream failed: comand was \"" << command << "\"");
      _exit(127);
  }
  /*
   * parent interact with the result
   */
  // fprintf (stderr, "proc_open: parent pid = %ld\n", long(pid));
  if (read_or_write != (READS | WRITES)) {
      close(child_desc);
      // fprintf (stderr, "proc_open: set fileno=%d\n", parent_desc);
      fp->fileno  = parent_desc;
      fp->fileno1 = 0;
  } else {
      close(p_desc  [1]);
      close(p_desc1 [0]);
      fp->fileno  = p_desc  [0];
      fp->fileno1 = p_desc1 [1];
  }
  /* link into proc_file_chain. */
  fp->next = proc_file_chain;
  proc_file_chain = fp;

  // fprintf (stderr, "proc_open: end parent.\n");
  return fp;
}
int
proc_close (struct procbuf_c *fp)
{
  // trace_macro ("> proc_close");
  if (!fp) {
    // trace_macro ("< proc_close: null proc pointer");
    return -1;
  }
  /* This is not name-space clean. FIXME! */
  int wstatus;
  struct procbuf_c **ptr = &proc_file_chain;
  pid_t wait_pid;
  int status = -1;
  
  /* Unlink from proc_file_chain. */
  for ( ; *ptr != 0; ptr = &(*ptr)->next)
    {
      if (*ptr == fp)
	{
          // trace_macro ("  proc_close: unlink pid = "<< fp -> pid);
	  *ptr = (*ptr)->next;
	  status = 0;
	  break;
	}
    }
  if (status < 0) {
    // trace_macro ("< proc_close: not-found-pid.");
    return -1;
  }
  if (close(fp->fileno) < 0) {
    // trace_macro ("  proc_close: close(fileno) failed; continue");
  } else {
    // trace_macro ("  proc_close: close(fileno=" << fp->fileno << ") successfull.");
  } 
  if (fp->fileno1) {
    if (fp->fileno1 && close(fp->fileno1) < 0) {
      // trace_macro ("  proc_close: close(fileno1) failed; continue");
    } else {
      // trace_macro ("  proc_close: close(fileno1=" << fp->fileno1 << ") successfull.");
    }
  }
  /* POSIX.2 Rationale:  "Some historical implementations either block
     or ignore the signals SIGINT, SIGQUIT, and SIGHUP while waiting
     for the child process to terminate.  Since this behavior is not
     described in POSIX.2, such implementations are not conforming." */
    
  // trace_macro ("  proc_close: now wait for child pid to stop...");
  do
    {
      wait_pid = waitpid (fp->pid, &wstatus, 0);
    } while (wait_pid == -1 && errno == EINTR);
  // trace_macro ("  proc_close: wait_pid = " << wait_pid);
  if (wait_pid == -1) {
    // trace_macro ("< proc_close: wait_pid failed.");
    return -1;
  }
  // trace_macro ("< proc_close: wait_pid successfull, wstatus = " << wstatus);
  return wstatus;
}

#define _MAGIC 0xFBAD0000 /* Magic number */

void
procbuf_c_init (struct procbuf_c *fp)
{
  fp -> fileno = 0;
  fp -> fileno1 = 0;
  fp -> pid = 0;
  fp -> next = 0;
}

#ifdef TODO
struct procbuf_c *
_popen (
    const char *command,
    const char *mode)
{
  struct procbuf_c *fpx = new procbuf_c;
  FILE *fp = (FILE*)fpx;
  if (fp == 0)
    return 0;
  proc_init(fp, 0);
  JUMPS(fp) = &proc_jumps;
  file_init(fp);
  if (proc_open (fp, command, mode) != 0)
    return fp;
  delete fpx;
  return 0;
}
struct jump_t proc_jumps = {
  JUMP_INIT_DUMMY,
  JUMP_INIT(finish, file_finish),
  JUMP_INIT(overflow, file_overflow),
  JUMP_INIT(underflow, file_underflow),
  JUMP_INIT(uflow, default_uflow),
  JUMP_INIT(pbackfail, default_pbackfail),
  JUMP_INIT(xsputn, file_xsputn),
  JUMP_INIT(xsgetn, default_xsgetn),
  JUMP_INIT(seekoff, file_seekoff),
  JUMP_INIT(seekpos, default_seekpos),
  JUMP_INIT(setbuf, file_setbuf),
  JUMP_INIT(sync, file_sync),
  JUMP_INIT(doallocate, file_doallocate),
  JUMP_INIT(read, file_read),
  JUMP_INIT(write, file_write),
  JUMP_INIT(seek, file_seek),
  JUMP_INIT(close, proc_close),
  JUMP_INIT(stat, file_stat)
};

#endif /* TODO */

