///
/// This file is part of Rheolef.
///
/// Copyright (C) 2000-2009 Pierre Saramito <Pierre.Saramito@imag.fr>
///
/// Rheolef is free software; you can redistribute it and/or modify
/// it under the terms of the GNU General Public License as published by
/// the Free Software Foundation; either version 2 of the License, or
/// (at your option) any later version.
///
/// Rheolef is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/// GNU General Public License for more details.
///
/// You should have received a copy of the GNU General Public License
/// along with Rheolef; if not, write to the Free Software
/// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
///
/// =========================================================================
#include "rheolef/pstream.h"
#include "rheolef/procbuf_c.h"

// ================= [  fileno  ] ==============================
/*
               +---------------------------------------------+
               |        ios_base            stdio equivalent |
               |        Value(s)                             |
               +---------------------------------------------+
               |in                          "r"              |
               |out | trunc                 "w"              |
               |out | app                   "a"              |
               |in | out                    "r+"             |
               |in | binary                 "rb"             |
               |out | trunc | binary        "wb"             |
               |out | app | binary          "ab"             |
               |in | out                    "r+              |
               |in | out | trunc            "w+"             |
               |in | out | app              "a+"             |
               |in | out | binary           "r+b"            |
               |in | out | trunc | binary   "w+b"            |
               |in | out | app | binary     "a+b"            |
               +---------------------------------------------+
*/
static 
int
get_fileno(
    const char *command, 
    int mode, 
    int prot, 
    struct procbuf_c** ptr_pp)
{
    *ptr_pp = new_macro(procbuf_c);
    procbuf_c_init(*ptr_pp);

    if (!command || strlen(command) == 0 || strcmp (command, "/bin/false") == 0) {
      return 0;
    }
    char* c_mode = 0;
    if ((mode & ios::in) && (mode & ios::out)) {
        c_mode = "r+";
    } else if (mode & ios::in) {
        c_mode = "r";
    } else if (mode & ios::out) {
        c_mode = "w";
    }
    if (! proc_open(*ptr_pp, command, c_mode) ) {
        return 0;
    }
    return (*ptr_pp) -> fileno;
}
static
FILE*
make_file_desc(int fileno, int mode)
{
    if (!fileno) return 0;
    const char* c_mode = ((mode & ios::in) ? "r" : "w");
    return fdopen(fileno, c_mode);
}
static
FILE*
get_file_desc(
    const char *name, 
    int mode, 
    int prot, 
    struct procbuf_c** ptr_pp)
{
    return 
    make_file_desc(get_fileno(name, mode, prot, ptr_pp), mode);
}
// ================== [ procbuf ] ==============================
procbuf::procbuf()
 : filebuf(), _pp(0), _mode(0)
{
}
// only used by bidirectional pstream (no work at all !): 
procbuf::procbuf(const procbuf& pbuf, int mode)
#ifdef _RHEOLEF_HAVE_FILEBUF_INT
   
   // GNU C++ v2 has file-number filebuf constructor
 : filebuf(pbuf._pp->fileno1), _pp(0), _mode(mode)

#elif _RHEOLEF_HAVE_FILEBUF_FILE
 
   // KAI C++ has file-descriptor filebuf constructor
 : filebuf(make_file_desc (pbuf._pp->fileno1, mode)), _pp(0), _mode(mode)

#elif _RHEOLEF_HAVE_FILEBUF_FILE_MODE
 
   // GNU C++ v3 has file-descriptor filebuf constructor with openmode
 : filebuf(make_file_desc (pbuf._pp->fileno1, mode), std::ios::openmode(mode)), _pp(0), _mode(mode)

#else
   have_neither_filebuf_fileno_nor_filebuf_file_constructor;
#endif // _RHEOLEF_HAVE_FILEBUF_INT
{
}
procbuf::procbuf(const char *c_name, int mode, int prot)
#ifdef _RHEOLEF_HAVE_FILEBUF_INT
   
   // GNU C++ has file-number filebuf constructor
 : filebuf(get_fileno(c_name, mode, prot, &_pp)), _mode(mode)

#elif _RHEOLEF_HAVE_FILEBUF_FILE
 
   // KAI C++ has file-descriptor filebuf constructor
 : filebuf(get_file_desc(c_name, mode, prot, &_pp)), _mode(mode)

#elif _RHEOLEF_HAVE_FILEBUF_FILE_MODE
 
   // GNU C++ v3 has file-descriptor filebuf constructor with openmode
 : filebuf(get_file_desc(c_name, mode, prot, &_pp), std::ios::openmode(mode)), _mode(mode)

#else
   have_neither_filebuf_fileno_nor_filebuf_file_constructor;
#endif // _RHEOLEF_HAVE_FILEBUF_INT
{
}
// I hope filebuf::attach(int) or filebuf::attach(FILE*) exists...
void
procbuf::open(const char *c_name, int mode, int prot)
{
    close();
    int fileno = get_fileno(c_name, mode, prot, &_pp);
    _mode = mode;
#ifdef _RHEOLEF_HAVE_FILEBUF_INT
    attach(fileno);
#elif defined(_RHEOLEF_HAVE_FILEBUF_FILE)
    FILE* file_desc = make_file_desc(fileno,mode);
    attach(file_desc);
#else
   // have_no_filebuf_attach;
   fatal_macro ("have no filebuf::attach(FILE*)");
#endif // _RHEOLEF_HAVE_FILEBUF_INT
}
// cstor used by bidirectional pstream
procbuf*
procbuf::close()
{
    // flush buffer and close
    filebuf::close();
    if (_pp) {
        proc_close(_pp);
    }
    return this;
}
procbuf::~procbuf()
{
    close();
    // stop child pid and close pipe
    if (_pp) {
        delete_macro(_pp);
        _pp = 0;
    }
}
bool
procbuf::is_open() const
{
    return _pp != 0 && _pp-> fileno != 0;
}
// ================= [ ipstream ] ==============================
ipstream::ipstream()
 : istream(ptr_pbuf = new_macro(procbuf()))
{
    setstate (ios::badbit);
}
ipstream::ipstream(const char *c_name, int mode, int prot)
 : istream(ptr_pbuf = new_macro(procbuf(c_name, mode, prot)))
{
    if (!ptr_pbuf || !ptr_pbuf -> is_open())
        setstate (ios::badbit);
}
// TODO: clean ::set_pbuf that requieres non-portable ios::rdbuf(procbuf*)
#if defined(_RHEOLEF_HAVE_ISTREAM_RDBUF) // KAI & GNU C++
void
ipstream::set_pbuf(procbuf* ptr_pbuf)
{
   rdbuf(ptr_pbuf);
}
void
opstream::set_pbuf(procbuf* ptr_pbuf)
{
   rdbuf(ptr_pbuf);
}
#elif defined(_RHEOLEF_HAVE_IOS_BP) // CRAY C++
void
ipstream::set_pbuf(procbuf* ptr_pbuf)
{
   bp = (streambuf*)(ptr_pbuf);
}
void
opstream::set_pbuf(procbuf* ptr_pbuf)
{
   bp = (streambuf*)(ptr_pbuf);
}
#else
void
ipstream::set_pbuf(procbuf* ptr_pbuf)
{
   fatal_macro("do not know how to access to ios::rdbuf() with this compiler -- sorry");
}
void
opstream::set_pbuf(procbuf* ptr_pbuf)
{
   fatal_macro("do not know how to access to ios::rdbuf() with this compiler -- sorry");
}
#endif // _RHEOLEF_HAVE_ISTREAM_RDBUF
void
ipstream::open(const char *name, int mode, int prot)
{
#ifdef TO_CLEAN
   // destroy possible procbuf
   if (ptr_pbuf) delete_macro(ptr_pbuf);

   // build a new procbuf
   ptr_pbuf = new_macro(procbuf(name, mode, prot));

   // link istream to this procbuf
   set_pbuf(ptr_pbuf);
#endif // TO_CLEAN

   // new strategy with procbuf::open(...)
   ptr_pbuf -> open(name, mode, prot);
    
   if (!ptr_pbuf -> is_open())
        setstate (ios::badbit);
}
void
ipstream::close()
{
#ifdef TO_CLEAN
    // disconnect istream from procbuf
    set_pbuf(0);

    // destroy procbuf
    if (ptr_pbuf) delete_macro(ptr_pbuf);
    ptr_pbuf = 0;
#endif // TO_CLEAN

   // new strategy with procbuf::close(...)
   ptr_pbuf -> close();
}
ipstream::~ipstream()
{
    ipstream::close();
}
// ================= [ opstream ] ==============================
opstream::opstream()
 : ostream(ptr_pbuf = new_macro(procbuf()))
{
}
opstream::opstream(const char *c_name, int mode, int prot)
 : ostream(ptr_pbuf = new_macro(procbuf(c_name, mode, prot)))
{
}
void
opstream::open(const char *name, int mode, int prot)
{
#ifdef TO_CLEAN
   // destroy possible procbuf
   if (ptr_pbuf) delete_macro(ptr_pbuf);

   // build a new procbuf
   ptr_pbuf = new_macro(procbuf(name, mode, prot));

   // link ostream to this procbuf
   // using ostream::rdbuf(streambuf*);
   set_pbuf(ptr_pbuf);
#endif // TO_CLEAN

   // new strategy with procbuf::open(...)
   ptr_pbuf -> open(name, mode, prot);
    
   if (!ptr_pbuf || !ptr_pbuf -> is_open())
        setstate (ios::badbit);
}
void
opstream::close()
{
#ifdef TO_CLEAN
    // disconnect ostream from procbuf
    set_pbuf(0);

    // destroy procbuf
    if (ptr_pbuf) delete_macro(ptr_pbuf);
    ptr_pbuf = 0;
#endif // TO_CLEAN

   // new strategy with procbuf::close(...)
   ptr_pbuf -> close();
}
opstream::~opstream()
{
    opstream::close();
}
// ================= [ pstream ] ==============================
#ifdef TODO
pstream::pstream()
 : istream(ptr_pbuf0 = new_macro(procbuf())),
   ostream(ptr_pbuf1 = new_macro(procbuf()))
{
}
pstream::pstream(const char *c_name, int mode, int prot)
 : istream(ptr_pbuf0 = new_macro(procbuf(c_name, ios::in|ios::out, prot))),
   ostream(ptr_pbuf1 = new_macro(procbuf(*ptr_pbuf0, ios::out)))
{
   istream::setf(ios::unitbuf);
   ostream::setf(ios::unitbuf);
   //istream::tie(&ostream(*this));
}
void
pstream::open(const char *name, int mode, int prot)
{
   // destroy possible procbufs
   if (ptr_pbuf0) delete_macro(ptr_pbuf0);
   if (ptr_pbuf1) delete_macro(ptr_pbuf1);

   // build new procbufs
   ptr_pbuf0 = new_macro(procbuf(name, ios::in|ios::out, prot));
   ptr_pbuf1 = new_macro(procbuf(*ptr_pbuf0, ios::out));

   // link istream and ostream to these procbufs
   istream::rdbuf (ptr_pbuf0);
   ostream::rdbuf (ptr_pbuf1);
   istream::setf(ios::unitbuf);
   ostream::setf(ios::unitbuf);
   //istream::tie(&ostream(*this));
}
void
pstream::close()
{
    // disconnect [i,o]stream from procbufs
    istream::rdbuf (0);
    ostream::rdbuf (0);

    // destroy procbufs
    if (ptr_pbuf0) delete_macro(ptr_pbuf0);
    ptr_pbuf0 = 0;
    if (ptr_pbuf1) delete_macro(ptr_pbuf1);
    ptr_pbuf1 = 0;
}
pstream::~pstream()
{
    pstream::close();
}
#endif // TODO
